@@ -3,8 +3,8 @@ import json
 from datetime import datetime
 from urllib.parse import quote
 
+import click
 import humanize
-import requests
 import rich
 import typer
 from rich.console import Console
@@ -13,18 +13,266 @@ from rich.table import Table
 from rich.theme import Theme
 from typer import Argument, Option
 
-# We instantiate a typer app and a rich console for later use
-app = typer.Typer(add_completion=False)
-console = Console(theme=Theme({"markdown.link": "#6088ff"}))
-# We will use these headers for every request
-headers = {"User-Agent": "wasi_master/pypi_cli"}
+try:
+    import click_help_colors
+    from click_help_colors import HelpColorsGroup, HelpColorsCommand
+except ImportError:
+    click_help_colors = None
+
+
+try:
+    from requests_cache.session import CachedSession
+except ImportError:
+    from requests import Session
+
+    session = Session()
+    session.headers.update({"User-Agent": "wasi_master/pypi_cli", "Accept": "application/json"})
+else:
+    import os.path  # pylint: disable=import-outside-toplevel
+
+    cache_path = os.path.join(os.path.dirname(__file__), "cache", "requests")
+    session = CachedSession(
+        cache_path,
+        backend="sqlite",
+        urls_expire_after={
+            **dict.fromkeys(
+                ["https://pypi.org/simple", "https://pypi.org/stats", "https://api.github.com/repos/*/readme"], 86400
+            ),
+            "https://pypi.org/pypi": 10800,
+            "https://pypi.org/search": 3600,
+            "https://pypi.org/rss": 60,
+            "https://pypistats.org/api/packages/": 21600,
+        },
+        headers={"User-Agent": "wasi_master/pypi_cli"},
+    )
+
+try:
+    import lxml
+except ImportError:
+    lxml = None
+
+
+def __color_error_message():
+    """Override click.UsageError.show to show colored output"""
+    from click._compat import get_text_stderr  # pylint: disable=import-outside-toplevel
+    from rich.markup import escape  # pylint: disable=import-outside-toplevel
+
+    def show(self, file=None):
+        if file is None:
+            file = get_text_stderr()
+        hint = ""
+        if self.ctx is not None and self.ctx.command.get_help_option(self.ctx) is not None:
+            hint = f"[magenta]Try '[blue]{self.ctx.command_path} [bold]{self.ctx.help_option_names[-1]}[/bold][/blue]' or visit [cyan]https://wasi-master.github.io/pypi-command-line/usage#{self.ctx.command.name.replace('-', '')}[/cyan] for help.[/magenta]"
+            hint = f"{hint}\n"
+        if self.ctx is not None:
+            console.print("[yellow]:disappointed_relieved: You did not do this properly[/]")
+            console.print(
+                f"{(self.ctx.get_usage().replace('...', '').replace('Usage: ', '[green]Usage: [/]').replace('[OPTIONS]', '[bright_black]'+ escape('[OPTIONS…]') + '[/]').replace('[ARGS]', '[bright_black]'+ escape('[ARGS…]') + '[/]'))}\n{hint}"
+            )
+        console.print(f":exclamation: [bold][red]Error[/bold]: {self.format_message()}[/red]")
+        try:
+            import questionary  # pylint: disable=import-outside-toplevel
+            from questionary import Choice, Style  # pylint: disable=import-outside-toplevel
+        except ImportError:
+            pass
+        else:
+            style = Style(
+                [
+                    ("link", "cyan"),
+                    ("command", "blue"),
+                    ("cancel", "gray"),
+                ]
+            )
+            print("\n")
+            resp = questionary.select(
+                "What do you want to do",
+                choices=[
+                    Choice(
+                        [
+                            ("class:text", "Run '"),
+                            ("class:command", f"{self.ctx.command_path} {self.ctx.help_option_names[-1]}"),
+                            ("class:text", "'"),
+                        ],
+                        value=0,
+                    ),
+                    Choice(
+                        [
+                            ("class:text", "Open "),
+                            (
+                                "class:link",
+                                f"https://wasi-master.github.io/pypi-command-line/usage#{self.ctx.command.name.replace('-', '')}",
+                            ),
+                        ],
+                        value=1,
+                    ),
+                    Choice([("class:cancel", "Nothing")], value=2),
+                ],
+                use_shortcuts=True,
+                style=style,
+            ).ask()
+            if resp == 0:
+                import os  # pylint: disable=import-outside-toplevel
+
+                console.print(f"[blue]❯ [/]{self.ctx.command_path} {self.ctx.help_option_names[-1]}")
+
+                os.system(f"{self.ctx.command_path} {self.ctx.help_option_names[-1]}")
+            elif resp == 1:
+                import webbrowser  # pylint: disable=import-outside-toplevel
+
+                webbrowser.open(
+                    f"https://wasi-master.github.io/pypi-command-line/usage#{self.ctx.command.name.replace('-', '')}"
+                )
+            else:
+                return
+
+    click.exceptions.UsageError.show = show
+
+
+if click_help_colors is not None:
+
+    class CustomHelpColorsGroup(HelpColorsGroup):
+        def __init__(self, *args, **kwargs) -> None:
+            super().__init__(*args, **kwargs)
+            self.help_headers_color = "red"
+            self.help_options_color = "yellow"
+            self.context_settings = {"help_option_names": ["-h", "--help"]}
+
+    class CustomHelpColorsCommand(HelpColorsCommand):
+        def __init__(self, *args, **kwargs) -> None:
+            super().__init__(*args, **kwargs)
+            self.help_headers_color = "red"
+            self.help_options_color = "yellow"
+
+    Group = CustomHelpColorsGroup
+    Command = CustomHelpColorsCommand
+else:
+    Group = click.Group
+    Command = click.Command
+
+
+class AliasedGroup(Group):
+    def get_command(self, ctx, cmd_name):
+        rv = click.Group.get_command(self, ctx, cmd_name)
+        if rv is not None:
+            return rv
+        alias_mapping = {**dict.fromkeys(["rtd", "docs", "documentation"], "read-the-docs"), "rs": "rsearch"}
+        if cmd_name in alias_mapping:
+            return click.Group.get_command(self, ctx, alias_mapping[cmd_name])
+        commands = self.list_commands(ctx)
+        matches = [x for x in commands if x.startswith(cmd_name)]
+        if not matches:
+            processor = lambda x: x.replace("-", "").lower()
+            try:
+                import rapidfuzz  # pylint: disable=import-outside-toplevel
+                
+                get_closest_match = lambda cmd: [
+                    i[0]
+                    for i in rapidfuzz.process.extract(
+                        cmd, commands, scorer=rapidfuzz.fuzz.WRatio, score_cutoff=50, processor=processor, limit=5
+                    )
+                ]
+            except ImportError:
+                try:
+                    import thefuzz.process  # pylint: disable=import-outside-toplevel
+                    import thefuzz.fuzz  # pylint: disable=import-outside-toplevel
+
+                    get_closest_match = lambda cmd: [
+                        i[0]
+                        for i in thefuzz.process.extractBests(
+                            cmd, commands, score_cutoff=50, processor=processor, limit=5
+                        )
+                    ]
+
+                except ImportError:
+                    import difflib  # pylint: disable=import-outside-toplevel
+
+                    get_closest_match = lambda cmd: difflib.get_close_matches(cmd, commands, n=5, cutoff=0.5) or [None]
+        if len(matches) == 0:
+            closest_matches = get_closest_match(cmd_name)
+            if not closest_matches:
+                # No match is more than 50% similar to the used name
+                return None
+
+            try:
+                import questionary  # pylint: disable=import-outside-toplevel
+            except ImportError:
+                console.print(
+                    f"""[cyan]ℹ️ Info:[/] Found invalid command '{cmd_name}', did you mean any of these: {', '.join(f"'[red]{match}[/]'" for match in closest_matches)}"""
+                )
+                raise typer.Exit()
+            else:
+                console.print(
+                    f"""[cyan]ℹ️ Info:[/] Found invalid command '{cmd_name}', closest matches: {', '.join(f"'[red]{match}[/]'" for match in closest_matches)}"""
+                )
+                resp = questionary.select(
+                    "Which one did you want to run?",
+                    choices=closest_matches,
+                    style=questionary.Style(
+                        [
+                            ("text", "red"),
+                            ("highlighted", "bg:ansibrightred"),
+                        ]
+                    ),
+                ).ask()
+
+            if not resp:
+                raise typer.Exit()
+            return click.Group.get_command(self, ctx, resp)
+        elif len(matches) == 1:
+            console.print(f"[cyan]ℹ️ Info:[/] Found shortened name '{cmd_name}', using '{matches[0]}'")
+            return click.Group.get_command(self, ctx, matches[0])
+        formatted_matches = ", ".join(sorted(f"[red]{match}[/]" for match in matches))
+        try:
+            import questionary
+        except ImportError:
+            ctx.fail(f"Found Too many matches for '{cmd_name}': {formatted_matches}")
+        else:
+            import difflib  # pylint: disable=import-outside-toplevel
+
+            console.print(f"[red]:warning: Attention:[/] Found Too many matches for '{cmd_name}': {formatted_matches}")
+            command = questionary.select(
+                "Select one to continue",
+                choices=difflib.get_close_matches(cmd_name, matches, cutoff=0.0),
+                style=questionary.Style(
+                    [
+                        ("text", "red"),
+                        ("highlighted", "bg:ansibrightred"),
+                    ]
+                ),
+            ).ask()
+            if not command:
+                raise typer.Exit()
+            return click.Group.get_command(self, ctx, command)
+
+    def resolve_command(self, ctx, args):
+        # always return the full command name
+        _, cmd, args = super().resolve_command(ctx, args)
+        return cmd.name, cmd, args
+
+
+class PypiTyper(typer.Typer):
+    """A custom subclassed version of typer.Typer to allow colored help"""
+
+    def __init__(self, *args, cls=AliasedGroup, **kwargs) -> None:
+        super().__init__(*args, cls=cls, **kwargs)
+
+    def command(self, *args, cls=Command, **kwargs) -> typer.Typer.command:
+        return super().command(*args, cls=cls, **kwargs)
+
+
+# We instantiate a cutom typer app
+app = PypiTyper()
+console = Console(theme=Theme({"markdown.link": "#6088ff"}), emoji=True, emoji_variant="emoji", tab_size=4)
+__color_error_message()  # makes the error messages colored
 
 
 class Package:
     """Represents a package gotten from scraping the search results."""
 
+    __slots__ = ("name", "version", "date", "released", "description")
+
     def __init__(self, soup):
-        """Instantiates a package
+        """Instantiate a package object gotten from scraping the search results.
 
         Parameters
         ----------
@@ -40,9 +288,9 @@ class Package:
         self.description = soup.find(class_="package-snippet__description").get_text()
 
 
-def utc_to_local(utc_dt):
+def utc_to_local(utc_dt, tzinfo):
     """Convert a datetime from utc to local time."""
-    return utc_dt.replace(tzinfo=timezone.utc).astimezone(tz=None)
+    return utc_dt.replace(tzinfo=tzinfo).astimezone(tz=None)
 
 
 def remove_dot_git(text):
@@ -69,22 +317,100 @@ def _format_classifiers(_classifiers: str):
     return output
 
 
-# 1. It first checks if the repository has a readme file. If it doesn't, it raises an exception.
-# 2. If it does, it gets the readme file's content and the path to the file.
-# 3. If the content is not found, it raises an exception.
-# 4. If the content is found, it returns the content and the path to the file.
+def load_cache():
+    import os  # pylint: disable=import-outside-toplevel
+
+    cache_file = os.path.join(os.path.dirname(__file__), "cache", "packages.txt")
+
+    try:
+        last_refreshed = os.path.getmtime(cache_file)
+    except FileNotFoundError:
+        return fill_cache(msg="Generating cache")
+    else:
+        import time  # pylint: disable=import-outside-toplevel
+
+        if time.time() - last_refreshed > 86400:
+            return fill_cache(msg="Cache is too old (>1d). Refreshing cache")
+        with open(cache_file, "r") as cache_file:
+            return cache_file.read().splitlines()
+
+
+def fill_cache(msg="Fetching cache"):
+    """Fill the cache with the packages."""
+    from rich.progress import Progress  # pylint: disable=import-outside-toplevel
+    import requests  # pylint: disable=import-outside-toplevel
+    import os  # pylint: disable=import-outside-toplevel
+
+    all_packages_url = "https://pypi.org/simple/"
+    cache_path = os.path.join(os.path.dirname(__file__), "cache")
+    if not os.path.exists(cache_path):
+        os.makedirs(cache_path)
+    cache_file = os.path.join(os.path.dirname(__file__), "cache", "packages.txt")
+
+    with Progress(transient=True) as progress:
+        response = requests.get(all_packages_url, stream=True)
+        response_data = ""
+        content_length = response.headers.get("content-length")
+        if content_length is not None:
+            total_length = int(content_length)
+            task = progress.add_task(msg, total=total_length)
+            downloaded = 0
+            for data in response.iter_content(chunk_size=32768):
+                downloaded += len(data)
+                response_data += data.decode("utf-8")
+                progress.advance(task, 32768)
+        else:
+            response_data = response.content.decode("utf-8")
+
+    import re  # pylint: disable=import-outside-toplevel
+
+    packages = re.findall(r"<a[^>]*>([^<]+)<\/a>", response_data)
+    with open(cache_file, "w", encoding="utf-8") as cache_file:
+        cache_file.write("\n".join(packages))
+    return packages
+
+
+def _refresh_cache():
+    with console.status("Getting current cache"):
+        old_cache = load_cache()
+    new_cache = fill_cache(msg="Fetching new cache")
+    changed = len(new_cache) - len(old_cache)
+    console.print(f"[yellow]:repeat: Updated the cache, number of new packages till last refresh:[/] [red]{changed}[/]")
+
+
+def _clear_cache():
+    try:
+        session.cache.clear()
+    except AttributeError:
+        pass
+    import os, shutil
+
+    folder = os.path.join(os.path.dirname(__file__), "cache")
+    for filename in os.listdir(folder):
+        file_path = os.path.join(folder, filename)
+        try:
+            if os.path.isfile(file_path):
+                os.remove(file_path)
+        except Exception as exc:
+            console.print(f"[red]:x: Failed to delete {file_path}. Reason: {exc}[/]")
+
+
 def _get_github_readme(repo):
-    readme = requests.get(f"https://api.github.com/repos/{repo}/readme").json()
+    readme = session.get(f"https://api.github.com/repos/{repo}/readme").json()
     if readme.get("message") == "Not Found":
-        console.print("[/]Could not find readme[/]")
+        console.print(f"[red]:x: Could not find readme for[/] [yellow]{repo}[/]")
         raise typer.Exit()
-    content = requests.get(f"https://raw.githubusercontent.com/{repo}/master/{readme['path']}")
+    content = session.get(f"https://raw.githubusercontent.com/{repo}/master/{readme['path']}")
     if content.status_code == 200:
         return content.text, readme["path"]
     return None, None
 
 
 def _format_xml_packages(url, title, pubmsg, _author, _link, *, split_title=False):
+    try:
+        import bs4  # pylint: disable=import-outside-toplevel
+    except ImportError:
+        bs4 = None
     table = Table(title=title, show_lines=True)
     table.add_column("Index", style="magenta", header_style="bold magenta")
     table.add_column("Name", style="green", header_style="bold green")
@@ -95,11 +421,17 @@ def _format_xml_packages(url, title, pubmsg, _author, _link, *, split_title=Fals
         table.add_column("Link", style="cyan", header_style="bold blue")
     table.add_column(pubmsg, style="yellow", header_style="bold yellow")
     with console.status("Fetching packages"):
-        response = requests.get(url, headers=headers)
-    soup = bs4.BeautifulSoup(response.text, "lxml-xml")
+        response = session.get(url)
+    if lxml and bs4:
+        soup = bs4.BeautifulSoup(response.text, "lxml-xml")
+    else:
+        import xml.etree.ElementTree as ET  # pylint: disable=import-outside-toplevel
+
+        soup = ET.fromstring(response.text)
+
     from datetime import timezone  # pylint: disable=import-outside-toplevel
 
-    for index, package in enumerate(soup.find_all("item"), 1):
+    for index, package in enumerate(soup.find_all("item") if lxml else soup.iter("item"), 1):
         title = package.find("title").text
         if split_title:
             title = title.split()[0]
@@ -107,9 +439,9 @@ def _format_xml_packages(url, title, pubmsg, _author, _link, *, split_title=Fals
         description = package.find("description")
         link = package.find("link").text
 
-        date = utc_to_local(datetime.strptime(package.find("pubDate").text, "%a, %d %b %Y %H:%M:%S GMT")).replace(
-            tzinfo=None
-        )
+        date = utc_to_local(
+            datetime.strptime(package.find("pubDate").text, "%a, %d %b %Y %H:%M:%S GMT"), timezone.utc
+        ).replace(tzinfo=None)
         if _link and _author:
             table.add_row(
                 f"{index}.",
@@ -135,29 +467,53 @@ def _format_xml_packages(url, title, pubmsg, _author, _link, *, split_title=Fals
                 humanize.naturaltime(date),
             )
     console.print(table)
+    if not lxml:
+        console.print(
+            "[bold yellow]:warning: WARNING: There is a known bug that occurs when lxml is not installed. It"
+            "doesn't show descriptions in some cases. Please install lxml using `pip install lxml`."
+        )
 
 
 @app.command()
-def desc(
+def description(
     package_name: str = Argument(..., help="Package to get the description for"),
-    force_github: bool = Argument(False, help="Forcefully get the description from github"),
+    force_github: bool = Option(False, help="Forcefully get the description from github"),
 ):
     """See the description for a package."""
     url = f"https://pypi.org/pypi/{quote(package_name)}/json"
     with console.status("Getting data from PyPI"):
-        response = requests.get(url, headers=headers)
+        response = session.get(url)
+
+    if response.status_code != 200:
+        if response.status_code == 404:
+            rich.print("[red]:no_entry_sign: Project not found[/]")
+        rich.print(f"[orange]:grey_exclamation: Some error occured. response code {response.status_code}[/]")
+        raise typer.Exit()
+
     parsed_data = json.loads(response.text)["info"]
     if force_github:
         import re  # pylint: disable=import-outside-toplevel
 
-        repo = re.findall(r"https://(www\.)?github\.com/([A-Za-z0-9_.-]{0,38}/[A-Za-z0-9_.-]{0,100})", str(parsed_data))
-        repo = repo[0][1] if repo else None
-        if not repo:
-            console.print("[red]I could not find a GitHub repository[/]")
+        repos = set(
+            re.findall(r"https://(?:www\.)?github\.com/([A-Za-z0-9_.-]{0,38}/[A-Za-z0-9_.-]{0,100})", str(parsed_data))
+        )
+        repos = {"https://github.com/wasi-master/package-x", "https://github.com/some-dude/some-other-repo"}
+        if len(repos) > 1:
+            console.print("[red]:warning: WARNING:[/] I found multiple github repos. ")
+            import questionary  # pylint: disable=import-outside-toplevel
+
+            repo = questionary.select(
+                "Please specify the repo you want to use.",
+                choices=[questionary.Choice([("cyan", r)]) for r in list(repos)],
+            ).ask()
+        elif len(repos) == 1:
+            repo = next(iter(repos))
+        else:
+            console.print("[red]:x: I could not find a GitHub repository[/]")
             raise typer.Exit()
         readme, filename = _get_github_readme(repo)
         if not readme or not filename:
-            console.print("[red]I could not find a readme inside the GitHub repository[/]")
+            console.print("[red]:x: I could not find a readme inside the GitHub repository[/]")
             raise typer.Exit()
         parsed_data["description"] = readme
         if filename.endswith((".md", ".md.txt")):
@@ -167,22 +523,40 @@ def desc(
         else:
             parsed_data["description_content_type"] = "text/markdown"
     if not parsed_data["description"] or parsed_data["description"] == "UNKNOWN":
-        console.print("[red]No description found on PyPI.[/]")
+        console.print("[red]:x: No description found on PyPI.[/]")
         import re  # pylint: disable=import-outside-toplevel
 
-        repo = re.findall(r"https://(www\.)?github\.com/([A-Za-z0-9_.-]{0,38}/[A-Za-z0-9_.-]{0,100})", str(parsed_data))
-        repo = repo[0][1] if repo else None
-        if repo:
-            import questionary  # pylint: disable=import-outside-toplevel
+        repos = set(
+            re.findall(r"https://(?:www\.)?github\.com/([A-Za-z0-9_.-]{0,38}/[A-Za-z0-9_.-]{0,100})", str(parsed_data))
+        )
+        if repos:
+            if len(repos) == 1:
+                repo = next(iter(repos))
+                console.print(f"[yellow]ℹ️ INFO:[/] However, I did find a github repo[/] https://github.com/{repo}.\n")
 
-            console.print(f"[yellow]However, I did find a github repo[/] https://github.com/{repo}.\n")
-            resp = questionary.confirm("Do you want to get the description from there?").ask()
-            if not resp:
-                console.print("[dim gray]Cancelled![/]")
-                raise typer.Exit()
+                try:
+                    import questionary  # pylint: disable=import-outside-toplevel
+                except ImportError:
+                    from rich.prompt import Confirm  # pylint: disable=import-outside-toplevel
+
+                    resp = Confirm.ask("Do you want to get the description from there?")
+                else:
+                    resp = questionary.confirm("Do you want to get the description from there?").ask()
+
+                if not resp:
+                    console.print("[dim gray]:ok: Cancelled![/]")
+                    raise typer.Exit()
+            elif len(repos) > 1:
+                console.print("[red]:warning: WARNING:[/] I did find some github repos. ")
+                import questionary  # pylint: disable=import-outside-toplevel
+
+                repo = questionary.select(
+                    "Please specify the repo you want to see the descripton from (Ctrl+C to cancel).",
+                    choices=[questionary.Choice([("cyan", r)]) for r in list(repos)],
+                ).ask()
             readme, filename = _get_github_readme(repo)
             if not readme or not filename:
-                console.print("[red]I could not find a readme inside the GitHub repository[/]")
+                console.print("[red]:x: I could not find a readme inside the GitHub repository[/]")
                 raise typer.Exit()
             parsed_data["description"] = readme
             if filename.endswith((".md", ".md.txt")):
@@ -191,6 +565,11 @@ def desc(
                 parsed_data["description_content_type"] = "text/x-rst"
             else:
                 parsed_data["description_content_type"] = "text/markdown"
+        else:
+            console.print(
+                "[red]:x: The PyPI page doesn't have a description nor a GitHub repository that I could've used[/]"
+            )
+            raise typer.Exit()
 
     if parsed_data["description_content_type"] == "text/markdown":
         from rich.markdown import Markdown  # pylint: disable=import-outside-toplevel
@@ -209,12 +588,10 @@ def desc(
 
 @app.command()
 def new_packages(
-    _author: bool = Option(False, metavar="author", help="Show the project author of not"),
-    _link: bool = Option(True, metavar="link", help="Show the project link of not"),
+    _author: bool = Option(False, metavar="author", help="Show the project author or not"),
+    _link: bool = Option(True, metavar="link", help="Show the project link or not"),
 ):
     """See the top 40 newly added packages."""
-    import bs4  # pylint: disable=import-outside-toplevel
-
     _format_xml_packages(
         "https://pypi.org/rss/packages.xml",
         "Newly Added Packages",
@@ -227,12 +604,10 @@ def new_packages(
 
 @app.command()
 def new_releases(
-    _author: bool = Option(False, metavar="author", help="Show the project author of not"),
-    _link: bool = Option(True, metavar="link", help="Show the project link of not"),
+    _author: bool = Option(False, metavar="author", help="Show the project author or not"),
+    _link: bool = Option(True, metavar="link", help="Show the project link or not"),
 ):
     """See the top 100 newly updated packages."""
-    import bs4  # pylint: disable=import-outside-toplevel
-
     _format_xml_packages(
         "https://pypi.org/rss/updates.xml",
         "Newly Released Packages",
@@ -249,7 +624,7 @@ def largest_files():
     headers = {"User-Agent": "wasi_master/pypi_cli", "Accept": "application/json"}
     url = "https://pypi.org/stats/"
     with console.status("Loading largest files..."):
-        response = requests.get(url, headers=headers)
+        response = session.get(url)
         data = json.loads(response.text)
     packages = data["top_packages"]
     packages = dict(sorted(packages.items(), key=lambda i: i[1]["size"], reverse=True))
@@ -279,7 +654,7 @@ def largest_files():
 
 @app.command()
 def search(
-    name: str = Argument(..., help="The name of the package to search for"),
+    package_name: str = Argument(..., help="The name of the package to search for"),
     page: int = Option(1, min=1, max=500, help="The page of the search results to show."),
     # classifier: List[str] = Option(
     #     None, help="Can be used multiple times to specify a list of classifiers to filter the results."
@@ -288,22 +663,22 @@ def search(
     """Search for a package on PyPI."""
     url = "https://pypi.org/search/"
     parameters = {
-        "q": name,
+        "q": package_name,
         "page": page,
     }
     # if classifier:
     #     parameters["c"] = classifier
-    with console.status(f"Searching for {name}..."):
-        response = requests.get(url, headers=headers, params=parameters)
+    with console.status(f"Searching for {package_name}..."):
+        response = session.get(url, params=parameters)
 
     if response.status_code == 404:
-        console.print("[bold]The specified page doesn't exist[/]")
+        console.print("[bold:no_entry_sign: The specified page doesn't exist[/]")
         raise typer.Exit()
 
     with console.status("Parsing data..."):
         import bs4  # pylint: disable=import-outside-toplevel
 
-        soup = bs4.BeautifulSoup(response.text, "lxml")
+        soup = bs4.BeautifulSoup(response.text, "lxml" if lxml else "html.parser")
         result_list = soup.find(attrs={"aria-label": "Search results"}, class_="unstyled")
         if not result_list:
             comment = soup.select(
@@ -322,7 +697,7 @@ def search(
 
     table = Table(
         show_header=True,
-        title=f"[bold]Search results for {name}[/]",
+        title=f"[bold]Search results for {package_name}[/]",
         show_lines=True,
         caption=f"Page {page} of {amount_of_pages}",
     )
@@ -355,7 +730,14 @@ def releases(
     """
     url = f"https://pypi.org/pypi/{quote(package_name)}/json"
     with console.status("Getting data from PyPI"):
-        response = requests.get(url, headers=headers)
+        response = session.get(url)
+
+    if response.status_code != 200:
+        if response.status_code == 404:
+            rich.print("[red]:no_entry_sign: Project not found[/]")
+        rich.print(f"[orange]:grey_exclamation: Some error occured. response code {response.status_code}[/]")
+        raise typer.Exit()
+
     parsed_data = json.loads(response.text)
 
     table = Table()
@@ -392,9 +774,9 @@ def releases(
 
 
 @app.command()
-def info(
+def information(
     package_name: str = Argument(...),
-    version: str = Option(None),
+    version: str = Option(None, help="The version of the package to show info for"),
     show_classifiers: bool = Option(False, metavar="classifiers", help="Show the classifiers"),
     hide_project_urls: bool = Option(False, metavar="project_urls", help="Hide the project urls"),
     hide_requirements: bool = Option(False, metavar="requirements", help="Hide the requirements"),
@@ -405,12 +787,12 @@ def info(
     """See the information about a package."""
     url = f"https://pypi.org/pypi/{quote(package_name)}{f'/{version}' if version else ''}/json"
     with console.status("Getting data from PyPI"):
-        response = requests.get(url, headers=headers)
+        response = session.get(url)
 
     if response.status_code != 200:
         if response.status_code == 404:
-            rich.print("[red]Project not found[/]")
-        rich.print(f"[orange]Some error occured. response code {response.status_code}[/]")
+            rich.print("[red]:no_entry_sign: Project not found[/]")
+        rich.print(f"[orange]:grey_exclamation: Some error occured. response code {response.status_code}[/]")
         raise typer.Exit()
 
     parsed_data = json.loads(response.text)
@@ -419,7 +801,10 @@ def info(
     releases = parsed_data["releases"]
     urls = parsed_data["urls"]
 
-    from packaging.version import parse as parse_version  # pylint:disable=import-outside-toplevel
+    try:
+        from packaging.version import parse as parse_version  # pylint:disable=import-outside-toplevel
+    except ImportError:
+        from distutils.version import LooseVersion as parse_version  # pylint:disable=import-outside-toplevel
 
     # HACK: should use fromisotime
     release_time = datetime.strptime(urls[-1]["upload_time_iso_8601"], "%Y-%m-%dT%H:%M:%S.%fZ")
@@ -464,7 +849,7 @@ def info(
         if repo:
             url = f"https://api.github.com/repos/{quote(repo)}"
             with console.status("Getting data from GitHub"):
-                resp = requests.get(url, headers=headers)
+                resp = session.get(url)
             github_data = json.loads(resp.text)
             if github_data.get("message") and github_data["message"] == "Not Found":
                 metadata.add_row(
@@ -495,7 +880,7 @@ def info(
     if not hide_stats:
         stats_url = f"https://pypistats.org/api/packages/{package_name}/recent"
         with console.status("Getting statistics from PyPI Stats"):
-            r = requests.get(stats_url)
+            r = session.get(stats_url)
         try:
             parsed_stats = json.loads(r.text)
             assert isinstance(parsed_stats, dict)
@@ -562,18 +947,16 @@ def info(
     console.print(metadata)
 
 
-@app.command("rsearch")
+@app.command()
 def regex_search(
     regex: str = Argument(..., help="The regular expression to search with"),
     compact: bool = Option(False, help="Compact formatting"),
 ) -> None:
     """Search for packages that match the regular expression."""
-    all_packages_url = "https://pypi.org/simple/"
-    with console.status("Fetching packages"):
-        response = requests.get(all_packages_url)
+    packages = load_cache()
+
     import re  # pylint: disable=import-outside-toplevel
 
-    packages = re.findall(r"<a[^>]*>([^<]+)<\/a>", response.text)
     # We compile the regex because it's twice as fast (https://imgur.com/a/MoUyEMg)
     _regex = re.compile(regex)
     if compact:
@@ -598,13 +981,17 @@ def regex_search(
                 )
         console.print(table)
         if table.row_count > 50:
-            console.print("[yellow]There are more than 50 matches, consider using the --compact flag")
+            console.print(
+                "[yellow]:warning: WARNING:[/] There are more than 50 matches, consider using the --compact flag"
+            )
 
 
 @app.command()
-def rtfd(
+def read_the_docs(
     package_name: str = Argument(..., help="The name or link to the docs of the package to show the documentation for"),
-    query: str = Argument(..., help="The query you want to read the docs for"),
+    query: str = Argument(
+        None, help="The query you want to read the docs for, if not passed goes to the main docs page"
+    ),
 ):
     """Search the documentation for an item of a package."""
     import webbrowser  # pylint: disable=import-outside-toplevel
@@ -672,25 +1059,34 @@ def rtfd(
             if resp:
                 url = f"https://pypi.org/pypi/{quote(package_name)}/json"
                 with console.status("Getting data from PyPI"):
-                    response = requests.get(url, headers=headers)
+                    response = session.get(url)
 
                 if response.status_code != 200:
                     if response.status_code == 404:
-                        rich.print("[red]Project not found[/]")
-                    rich.print(f"[orange]Some error occured. response code {response.status_code}[/]")
+                        rich.print("[red]:no_entry_sign: Project not found[/]")
+                    rich.print(
+                        f"[orange]:grey_exclamation: Some error occured. response code {response.status_code}[/]"
+                    )
                     raise typer.Exit()
 
                 parsed_data = json.loads(response.text)
                 url = parsed_data["info"].get("project_urls", {}).get("Documentation", None)
                 if not url:
-                    console.print("[bold]Documentation url not found on PyPI[/]")
+                    console.print("[bold]:x: Documentation url not found on PyPI[/]")
                     raise typer.Exit()
                 else:
-                    url = url + "/search.html"
+                    import os.path  # pylint: disable=import-outside-toplevel
+
+                    if "readthedocs.io" in url:
+                        url = os.path.join(url, "en/stable/")
+                    url = os.path.join(url, "search.html")
             else:
-                console.print("[dim grey]Cancelled![/]")
+                console.print("[dim grey]:ok: Cancelled![/]")
                 raise typer.Exit()
 
+    if not query:
+        webbrowser.open(url.replace("search.html", ""))
+        raise typer.Exit()
     search_page = url + "?q=" + quote(query)
     webbrowser.open(search_page)
 
@@ -710,13 +1106,14 @@ def browse(package_name: str = Argument(...)):
     )
 
     url = f"https://pypi.org/pypi/{quote(package_name)}/json"
+
     with console.status("Getting data from PyPI"):
-        response = requests.get(url, headers=headers)
+        response = session.get(url)
 
     if response.status_code != 200:
         if response.status_code == 404:
-            rich.print("[red]Project not found[/]")
-        rich.print(f"[orange]Some error occured. response code {response.status_code}[/]")
+            rich.print("[red]:no_entry_sign: Project not found[/]")
+        rich.print(f"[orange]:grey_exclamation: Some error occured. response code {response.status_code}[/]")
         raise typer.Exit()
 
     parsed_data = json.loads(response.text)
@@ -743,6 +1140,66 @@ def browse(package_name: str = Argument(...)):
         webbrowser.open(answer)
 
 
+@app.command()
+def cache_refresh():
+    """Refresh the cache."""
+    _refresh_cache()
+
+
+@app.command()
+def cache_clear():
+    """Clear the cache."""
+    _clear_cache()
+
+
+@app.command()
+def cache_info():
+    """See information about the cache"""
+    import os.path  # pylint: disable=import-outside-toplevel
+
+    packages_cache = os.path.join(os.path.dirname(__file__), "cache", "packages.txt")
+    requests_cache = os.path.join(os.path.dirname(__file__), "cache", "requests.sqlite")
+    try:
+        packages_size = os.path.getsize(packages_cache)
+        packages_last_refreshed = os.path.getmtime(packages_cache)
+    except FileNotFoundError:
+        packages_size = None
+        packages_last_refreshed = None
+        console.print("[bold yellow]:no_entry_sign: Packages cache not available[/]")
+    try:
+        requests_size = os.path.getsize(requests_cache)
+    except FileNotFoundError:
+        requests_size = None
+        console.print("[bold yellow]:no_entry_sign: Requests cache not available[/]")
+        if not packages_size:
+            console.print("[bold red]:warning::no_entry_sign: No cache available![/]")
+            # If both the caches are unavailable, then we can't do anything
+            raise typer.Exit()
+
+    console.print(f"ℹ️ Packages cache size: {humanize.naturalsize(packages_size or 0, binary=True)}")
+
+    console.print(f"ℹ️ Requests cache size: {humanize.naturalsize(requests_size or 0, binary=True)}")
+    console.print(f"ℹ️ Requests cache size: {humanize.naturalsize(requests_size or 0, binary=True)}")
+    if packages_last_refreshed:
+        from datetime import datetime
+
+        console.print(
+            f"⏰ Requests cache last updated: {humanize.naturaltime(datetime.fromtimestamp(packages_last_refreshed))}"
+        )
+
+    if requests_size:
+        table = Table(title="All cached requests")
+        table.add_column("Index", style="dim magenta", header_style="bold magenta")
+        table.add_column("Link", style="cyan", header_style="bold cyan")
+        table.add_column("Created", style="green", header_style="bold green")
+        table.add_column("Expires", style="green", header_style="bold green")
+        for n, response in enumerate(session.cache.values()):
+            table.add_row(
+                f"{n}.", response.url, humanize.naturaltime(response.created_at), humanize.naturaltime(response.expires)
+            )
+        console.print(table)
+
+
 def run():
     """Run the CLI."""
     app()
